#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <sdsl/suffixarrays.hpp>
#include <sdsl/testutils.hpp>

#include <iostream>
#include <fstream>
#include <string>

using namespace std;
namespace po = boost::program_options;
namespace fs = boost::filesystem;

typedef sdsl::csa_sada_prac<sdsl::enc_vector_prac<sdsl::coder::elias_delta, 0>, 0> tCSAEd;

int main(int ac, char* av[]){
	const uint16_t version = 0, subversion = 1;
	uint32_t compression = 1;
	uint32_t psi_sample_dens = 128;
	uint32_t sa_sample_dens = 8;
	bool verbose = false;
	string input_file="";

	po::options_description hidden("Hidden options");
	hidden.add_options()
		 ("input-file", po::value<string>(&input_file)->default_value(""),"input file")
	;

	po::options_description desc("Allowed options");
	desc.add_options()
		 ("help", "produce help message")
		 ("compression,c", po::value<uint32_t>(&compression)->default_value(10), "set compresssion level")
		 ("verbose,v", po::value<bool>(&verbose)->default_value(false), "verbose mode")
	;
	
	po::variables_map vm;
	po::store(po::parse_command_line(ac, av, desc), vm);
	po::notify(vm);

	po::positional_options_description pos_option_desc;
	pos_option_desc.add("input-file", -1);
	po::store(po::command_line_parser(ac, av).options(hidden).positional(pos_option_desc).run(), vm);
	po::notify(vm);

	if(vm.count("help")){
		cout << desc << endl;
		return 1;
	}

	if(input_file == ""){
		cout << "idxz version "<<version<<"."<<subversion<<" - a compressed full-text index" << endl;
	    cout << "(c) Simon Gog 2008" <<endl;
		cout << "Please specify an input file." << endl;
		cout << "Usage: ./idxz [options] input-file" << endl << endl;
		cout << desc << endl;
		return 0;
	}

	if( verbose ){
		if(vm.count("compression")){
			cout << "Compression level was set to " <<
				   vm["compression"].as<uint32_t>() << "." << endl;
			sa_sample_dens = vm["compression"].as<uint32_t>();
		} else {
			cout << "Compression level was not set." << endl;
		}
	}
	
	char *text = NULL;
	uint64_t n=0;
	if( (n=sdsl::file::readText( input_file.c_str(), text )) > 0 ){
		tCSAEd csa;
		csa.set_psi_sample_dens(psi_sample_dens);
		csa.set_sample_dens(sa_sample_dens);
		csa = tCSAEd((const unsigned char*)text);
		sdsl::util::storeToFile(csa, (input_file+".csa").c_str() );
		delete [] text;
	}else{
		cerr << "Could not open file \""<< input_file << "\"" << endl;
	}	
	
}
