/*! \file bit_vector_handling.e.cpp
 *  \brief This is a test program for the class  sdsl::bit_vector.
 *
 */ 
#include "sdsl/int_vector.hpp"
#include <iostream>
#include <string>
#include <cassert>
#include <fstream>

using namespace std;
using namespace sdsl;

int main(){
	// Create a new bit_vector with 123 bits.
	bit_vector v(123);
	// Use util to set all bits to zero.
	sdsl::util::setZeroBits(v);
	v[122]=1;
	// Set bits to specific values.
	v[5]=v[8]=1; v[10]=1;
	cout<< "v[5]=v[8]=1; v[10]=1;" << endl;
	cout<< "v=" << v << endl;

	cout << "Set the bits of v to random bits:" << endl;	
	util::setRandomBits(v);

	bit_vector v2 = v;
	cout<<"Make a copy of v:\n"<<v2<<endl;
	v2[0]=v2[5]=v2[4]=v2[3]=v2[2]=v2[1]=1;
	cout<<"Output the modified copy v2 of v:\n"<<v2<<endl;
	cout<<"Output v:\n"<<v<<endl;
	
	const char* bv_file_name = "bitvector_data";

	cout << "Try to load a SDSBitVector from the file " << bv_file_name << "." << endl;
	if( !sdsl::util::loadFromFile(v, bv_file_name) ){
		cerr << "File "<< bv_file_name <<" does not exist. Creating new one..."<< endl;
		v.bit_resize(100000000);
		util::setRandomBits(v);
		if( !util::storeToFile(v, bv_file_name) )
			cerr << "ERROR: Could not create file "<< bv_file_name <<"." << endl;
	}else{
		cout<<"Loaded bit_vector from file "<< bv_file_name << "." << endl;
	}

	//Try vector of fixed int width.
	int_vector<64> a64(3);
	a64[0] = 911;
	a64[1] = 1024;
	a64[2] = 747;
	cout<< a64[0] << " " << a64[1] << " " << a64[2] <<endl;
	int_vector<32> a32(3);
	a32[0] = 911;
	a32[1] = 1024;
	a32[2] = 737;
	cout<< a32[0] << " " << a32[1] << " " << a32[2] <<endl;
}
