#include <sdsl/vectors.hpp>
#include <sdsl/testutils.hpp>
#include <vector>

using namespace sdsl;
using namespace std;

int main(){

	cout<< "measure the performance of int_vector<>, int_vector<64>, vector<uint64_t>, and " << endl;
	size_t n = 10000000, cnt=0;
	const size_t inc = 5;
	cout<<"size of data structures:  "<< 8*n << " bytes" <<endl;
	stop_watch stopWatch;
	{
		stopWatch.start();
		int_vector<>   i64(n,0,64); // create an int_vector with int width = 64, int width could be changed.
		stopWatch.stop();

		cout<<"int_vector<> build takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		stopWatch.start();
		for(size_t i=0, size = i64.size(); i<size;++i) i64[i] = i;
		stopWatch.stop();
		cout<<"int_vector<> initialization with values takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		cnt = 0;
		stopWatch.start();
		for(int_vector<>::iterator it=i64.begin(),end=i64.end(); it!=end; ++it, cnt+=inc) *it = cnt;
		stopWatch.stop();
		cout<<"int_vector<> initialization with values takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		cnt = 0;
		stopWatch.start();
		for(int_vector<>::const_iterator it=i64.begin(),end=i64.end(); it!=end; ++it, cnt+=inc)
		   if(*it != cnt)
			   assert(false);
		stopWatch.stop();
		cout<<"int_vector<> check takes "<<stopWatch.getUserTime()<<" ms"<<endl;
	}

	{
		stopWatch.start();
		int_vector<64>   i64(n); // create an int_vector with int width = 64, int width is fixed.
		stopWatch.stop();

		cout<<"int_vector<64> build "<<stopWatch.getUserTime()<<" ms"<<endl;
		stopWatch.start();
		for(size_t i=0, size = i64.size(); i<size;++i) i64[i] = i;
		stopWatch.stop();
		cout<<"int_vector<64> initialization with values takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		cnt = 0;
		stopWatch.start();
		for(int_vector<64>::iterator it=i64.begin(),end=i64.end(); it!=end; ++it, cnt+=inc) *it = cnt;
		stopWatch.stop();
		cout<<"int_vector<64> initialization with values takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		stopWatch.start();
		cnt = 0;
		for(int_vector<64>::const_iterator it=i64.begin(),end=i64.end(); it!=end; ++it, cnt+=inc)
		   if(*it != cnt)
			   assert(false);
		stopWatch.stop();
		cout<<"int_vector<> check takes "<<stopWatch.getUserTime()<<" ms"<<endl;

	}

	enc_vector<> ev;
	{
		stopWatch.start();
		vector<uint64_t> v64(n); // create a stl vector of 64 bit integers.
		stopWatch.stop();

		cout<<"vector<uint64_t> build takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		stopWatch.start();
		for(size_t i=0, size = v64.size(); i<size;++i) v64[i] = i;
		stopWatch.stop();
		cout<<"vector<uint64_t> initialization with values takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		cnt = 0;
		stopWatch.start();
		for(vector<uint64_t>::iterator it=v64.begin(),end=v64.end(); it!=end; ++it, cnt+=inc) *it = cnt;
		stopWatch.stop();
		cout<<"vector<uint64_t> initialization with values takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		cnt = 0;
		stopWatch.start();
		for(vector<uint64_t>::const_iterator it=v64.begin(),end=v64.end(); it!=end; ++it, cnt+=inc)
		   if(*it != cnt)
			   assert(false);
		stopWatch.stop();
		cout<<"int_vector<> check takes "<<stopWatch.getUserTime()<<" ms"<<endl;
		
		stopWatch.start();
		ev = v64;
		stopWatch.stop();
		cout<<"creation of enc_vector<> from v64 takes "<<stopWatch.getUserTime()<<" ms"<<endl;
	}
	cnt = 0;
	stopWatch.start();
	for(enc_vector<>::const_iterator it=ev.begin(), end = ev.end(); it!=end; ++it, cnt+=inc)
		if(*it != cnt)
			assert(false);
	stopWatch.stop();
	cout<<"enc_vector<> check takes "<<stopWatch.getUserTime()<<" ms"<<endl;
}
