#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/algorithms.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/testutils.hpp>
#include <algorithm> // for stl conform test 

using namespace boost::unit_test;
using namespace sdsl;
using std::vector;

BOOST_AUTO_TEST_SUITE( algorithm_performance_suite )

BOOST_AUTO_TEST_CASE( performance_test )
{
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE("Running performance test for algorithm");
	
	size_t s = 10000000;
	
	stop_watch stopWatch;
	
	{
		unsigned int size=10000000;
		vector<unsigned int> v(size), psv(size);
		for(size_t i=0; i<v.size(); ++i){
			v[i] = i;
		}
		std::random_shuffle(v.begin(), v.end());

		stopWatch.start();
		algorithm::calculate_psv(v, psv);
		stopWatch.stop();
		BOOST_MESSAGE("Calculating psv of size "<< size <<" for std::vector takes "<< stopWatch.getUserTime() << " ms." );

		stopWatch.start();
		BOOST_CHECK(algorithm::verify_psv(v, psv));
		stopWatch.stop();
		BOOST_MESSAGE("Verified psv of size "<< size  <<" in "<< stopWatch.getUserTime() <<" ms.");
	}

	{
		unsigned int size=10000000;
		vector<unsigned int> v(size), nsv(size);
		for(size_t i=0; i<v.size(); ++i){
			v[i] = i;
		}
		std::random_shuffle(v.begin(), v.end());

		stopWatch.start();
		algorithm::calculate_nsv(v, nsv);
		stopWatch.stop();
		BOOST_MESSAGE("Calculating nsv of size "<< size <<" for std::vector takes "<< stopWatch.getUserTime() << " ms." );

		stopWatch.start();
		BOOST_CHECK(algorithm::verify_nsv(v, nsv));
		stopWatch.stop();
		BOOST_MESSAGE("Verified nsv of size "<< size  <<" in "<< stopWatch.getUserTime() <<" ms.");
	}

	{
		vector<unsigned int> v(s), psv(s);
		for(size_t i=0; i<s; ++i){
			v[i] = i;
		}
		std::random_shuffle(v.begin(), v.end());

		stopWatch.start();
		algorithm::calculate_psv(v, psv);
		stopWatch.stop();
		BOOST_MESSAGE("Calculating psv for std::vector takes "<< stopWatch.getUserTime() << " ms." );
	}

	{
		int_vector<32> v(s), psv(s);
		for(size_t i=0; i<s; ++i){
			v[i] = i;
		}
		std::random_shuffle(v.begin(), v.end());

		stopWatch.start();
		algorithm::calculate_psv(v, psv);
		stopWatch.stop();
		BOOST_MESSAGE("Calculating psv for int_vector<32> takes "<< stopWatch.getUserTime() << " ms." );

	}
	{
		int_vector<32> v(s), psv(s);
		for(size_t i=0; i<s; ++i){
			v[i] = i;
		}
		std::random_shuffle(v.begin(), v.end());

		stopWatch.start();
		algorithm::calculate_psv2(v, psv);
		stopWatch.stop();
		BOOST_MESSAGE("Calculating psv2 for int_vector<32> takes "<< stopWatch.getUserTime() << " ms." );

	}
	{
		vector<unsigned int> v(s), psv(s);
		for(size_t i=0; i<s; ++i){
			v[i] = i;
		}
		std::random_shuffle(v.begin(), v.end());

		stopWatch.start();
		algorithm::calculate_psv2(v, psv);
		stopWatch.stop();
		BOOST_MESSAGE("Calculating psv2 for std::vector takes "<< stopWatch.getUserTime() << " ms." );

	}
}

BOOST_AUTO_TEST_SUITE_END()

/*test_suite*
init_unit_test_suite( int , char* [] )
{
    framework::master_test_suite().p_name.value = "SDSBitVector Test Suite";

    // with explicit registration we could specify a test case timeout
    framework::master_test_suite().add( BOOST_TEST_CASE( &operator_test ), 0 );

    return 0; 
}
*/
