#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/testutils.hpp>

using namespace boost::unit_test;
using namespace boost::test_tools;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( SDSMagic_performance_suite )

BOOST_AUTO_TEST_CASE( performance_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_MESSAGE( "Running performance test for bit_magic..." );
	BOOST_MESSAGE(" res = "<<bit_magic::b11CntS(3));
	
	stop_watch stopwatch;

	int32_t lg = 0, lg2 = 0;

	int_vector<64> v(100000000);
	util::setRandomBits(v);

	size_t size = v.size();
	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg+=bit_magic::b1Cnt(x);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" b1Cnt calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg2 += bit_magic::b1CntNaive(x);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" b1CntNaive calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );
	BOOST_CHECK(lg == lg2);

	lg = lg2 = 0;
	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg+=bit_magic::b11Cnt(x);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" b11Cnt calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg2+=bit_magic::b11CntS(x);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" b11CntS calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );
	BOOST_CHECK(lg == lg2);
	if(lg!=lg2){
		for(uint64_t x=0; x<size; ++x){
			if(bit_magic::b11Cnt(x)!=bit_magic::b11CntS(x)){
				BOOST_MESSAGE("x = "<<x<<" b11Cnt="<<bit_magic::b11Cnt(x)<<" b11CntS="<<bit_magic::b11CntS(x));
				break;
			}
		}
	}
	lg2 = 0;

	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg2+=bit_magic::b11CntNaive(x);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" b11CntNaive calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );
	BOOST_CHECK(lg==lg2);

	lg = lg2 = 0;
	stopwatch.start();
	for(uint64_t x=0; x<size/14; ++x){
		for(uint32_t i=1;i<=14;++i){
			lg+=bit_magic::i1BP(45036179514458794ULL,i);
		}
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" i1BP calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/((size/14)*14) << " ms" );

	stopwatch.start();
	for(uint64_t x=0; x<size/14; ++x){
		for(uint32_t i=1;i<=14;++i){
			lg2+=bit_magic::i1BPNaive(45036179514458794ULL,i);
		}
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" i1BPNaive calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/((size/14)*14) << " ms" );
	BOOST_CHECK(lg==lg2);

	lg2 = 0;
	stopwatch.start();
	for(uint64_t x=0; x<size/14; ++x){
		for(uint32_t i=1;i<=14;++i){
			lg2+=bit_magic::j1BP(45036179514458794ULL,i);
//			if (bit_magic::j1BP(45036179514458794ULL,i) != bit_magic::i1BP(45036179514458794ULL,i) ){
//				BOOST_MESSAGE(" i= "<<i<<" j1BP="<<bit_magic::j1BP(45036179514458794ULL,i)<<" i1BP="<<bit_magic::i1BP(45036179514458794ULL,i) );
//				BOOST_REQUIRE(false);
//			}
		}
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" j1BP calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/((size/14)*14) << " ms" );
	BOOST_CHECK(lg==lg2);

	lg=lg2=0;
	stopwatch.start();
	for(uint64_t x=0; x<size/32; ++x){
		for(uint32_t i=1;i<=32;++i){
			lg+=bit_magic::i11BP(0xFFFFFFFFFFFFFFFFULL,i);
		}
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" i11BP calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/((size/32)*32) << " ms" );

	stopwatch.start();
	for(uint64_t x=0; x<size/32; ++x){
		for(uint32_t i=1;i<=32;++i){
			lg2+=bit_magic::i11BPNaive(0xFFFFFFFFFFFFFFFFULL,i);
		}
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" i11BPNaive calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/((size/32)*32) << " ms" );
	BOOST_CHECK( lg==lg2 );


	BOOST_CHECK( bit_magic::l1BP(0)==0 );
	BOOST_CHECK( bit_magic::l1BP(1)==0 );
	BOOST_CHECK( bit_magic::r1BP(0)==0 );
	for(unsigned i=1; i<64; ++i){
		BOOST_CHECK( bit_magic::l1BP(1ULL << i) == i );
		BOOST_CHECK( bit_magic::r1BP(1ULL << i) == i );
	}
	lg = lg2 = 0;
	stopwatch.start();
	for(uint64_t i=0, x=0x3AAAAAAAAAAAAAAAULL; i<size; ++i){
		lg += bit_magic::l1BP(x);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" l1BP calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );



	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg += bit_magic::r1BP(x);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" r1BP calls "<< stopwatch.getUserTime() << "ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg+=bit_magic::l11BP(x);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" l11BP calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	uint16_t temp;
	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg+=bit_magic::max_excess(x, temp);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" max_excess calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg+=bit_magic::max_excess2(x, temp);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" max_excess2 calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg+=bit_magic::max_excess3(x, temp);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" max_excess3 calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	uint64_t temp2=0,temp3=0;
	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		bit_magic::max_byte_excesses2(x, temp2, temp3);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" max_byte_excesses calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );
	BOOST_MESSAGE(""<<temp2<<" "<<temp3);

	uint64_t temp4=0;
	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		bit_magic::min_max_byte_excesses(x, temp2, temp3, temp4);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" min_max_byte_excesses calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );
	BOOST_MESSAGE(""<<temp2<<" "<<temp3<<" "<<temp4);



	uint64_t sums;
	stopwatch.start();
	for(uint64_t x=0, y=9876543210; x<size; ++x, y=y*y-1){
		lg+=bit_magic::first_excess_position(y, 1, sums);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" first_excess_position calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	stopwatch.start();
	for(uint64_t x=0,y=9876543210; x<size; ++x, y=y*y-1){
		lg+=bit_magic::first_excess_position_naive(y, 1, sums);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" first_excess_position_naive calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	stopwatch.start();
	for(uint64_t x=0,y=9876543210; x<size; ++x, y=y*y-1){
		y &= 0xFFFFFFFF7FFFFFFFULL;
		lg+=bit_magic::find_open(y, 31);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" find_open calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );

	stopwatch.start();
	for(uint64_t x=0,y=9876543210; x<size; ++x, y=y*y-1){
		y &= 0xFFFFFFFF7FFFFFFFULL;
		lg+=bit_magic::find_open_naive(y, 31);
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" find_open_naive calls "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );





	stopwatch.start();
	for(uint64_t x=0; x<size; ++x){
		lg += x & 0x3F;
	}
	stopwatch.stop();
	BOOST_MESSAGE("Time for "<<size<<" assignments "<< stopwatch.getUserTime() << " ms");
	BOOST_MESSAGE("Per call: "<< stopwatch.getUserTime()/size << " ms" );
	BOOST_MESSAGE("checksum = "<<lg);

//	bit_magic::generate_pos_of_excess_val();
}

BOOST_AUTO_TEST_SUITE_END()
