#include <sdsl/cst_sada.hpp> 
#include <sdsl/cst_sct.hpp>
#include <sdsl/testutils.hpp>
#include <sdsl/suffixarrays.hpp>


#include <iostream>
#include <fstream>

#include <boost/filesystem.hpp>
#if BOOST_WORKAROUND( __GNUC__, < 3)
typedef boost::test_tools::output_test_stream onullstream_type;
#else
#include <boost/test/utils/nullstream.hpp>
typedef boost::onullstream onullstream_type;
#endif

using namespace sdsl;
using namespace boost::filesystem;
using namespace std;

typedef cst_sct<>			cst_type;
//typedef cst_sada<>		    cst_type;
//typedef cst_sada<cst_sada<>::csa_type, int_vector<0> >	    cst_type;
//typedef cst_sct<cst_sct<>::csa_type, int_vector<0> >	    cst_type;
//typedef cst_type::size_type size_type;
//typedef cst_type::node_type node_type;

template<class T>
struct cst_name_trait{
	static string file_name_suffix(){
		return ".cst";	
	} 
};

template<uint32_t sa_sample_dens, uint32_t psi_sample_dens, class Coder>
struct cst_name_trait<cst_sct<csa_sada_prac<enc_vector_prac<Coder, psi_sample_dens>, sa_sample_dens, sa_sample_dens> > >{
	static string file_name_suffix(){
		if( typeid(Coder) == typeid(coder::elias_delta) )
			return ".ed.sct.cst";
		return ".sct.cst";
	}
};

template<uint32_t sa_sample_dens, uint32_t psi_sample_dens, class Coder>
struct cst_name_trait<cst_sada<csa_sada_prac<enc_vector_prac<Coder, psi_sample_dens>, sa_sample_dens, sa_sample_dens> > >{
	static string file_name_suffix(){
		if( typeid(Coder) == typeid(coder::elias_delta) )
			return ".ed.sada.cst";
		return ".sada.cst";
	}
};

template<>
struct cst_name_trait<cst_sct<csa_uncompressed, int_vector<0> > >{
	static string file_name_suffix(){
		return ".allfix.sct.cst";
	}
};

template<>
struct cst_name_trait<cst_sada<csa_uncompressed, int_vector<0> > >{
	static string file_name_suffix(){
		return ".allfix.sada.cst";
	}
};

template<>
struct cst_name_trait<cst_sct<cst_sct<>::csa_type, int_vector<0> > >{
	static string file_name_suffix(){
		return ".lcpfix.sct.cst";
	}
};

template<>
struct cst_name_trait<cst_sada<cst_sada<>::csa_type, int_vector<0> > >{
	static string file_name_suffix(){
		return ".lcpfix.sada.cst";
	}
};

template <class CST>
bool create_or_load_cst(CST &cst, string cst_suffix, string test_case, string test_case_dir, string data_dir){
	typedef typename CST::csa_type CSA;
	if( !util::loadFromFile(cst, (data_dir+test_case+cst_suffix).c_str() ) ){
		char *text = NULL;
		uint64_t n=0;
		if( 0==(n=file::readText((test_case_dir+test_case).c_str(),text)) ){
			cerr << "#Could not read file "<<(test_case_dir+test_case)<<". Exiting. "<< endl;	
			return false;
		}	
		CSA csa;
		int_vector<32> sa;
		// if we already calculated the suffix array, laod it from disc
		if( !util::loadFromFile(sa, (data_dir+test_case+".sa").c_str() ) ){ 
			cerr<<"#sa already calculated"<<endl;
			sa.resize(n+1);
			algorithm::calculate_sa((const unsigned char*)text,n+1,sa);
			util::storeToFile(sa, ((data_dir+test_case)+".sa").c_str() );
		}
		csa = CSA(sa, (const unsigned char*)text);
		sa.resize(0); // clear sa
		cst = CST(csa, (const unsigned char*)text);
		util::storeToFile(cst, (data_dir+test_case+cst_suffix).c_str() );
		std::cerr<<"#free text"<<std::endl;
		delete [] text;
		std::cerr<<"#text was deleted"<<std::endl;
	}
	return true;
}

template <class CST>
void dfs(const CST &cst, typename CST::node_type v, vector<typename CST::node_type> &nodes){
	if(nodes.size()%1000000==0){
		std::cerr<<".";
	}
	nodes.push_back(v);
	typename CST::size_type i=1;
	typename CST::node_type temp = cst.root();
	while( (temp = cst.ith_child(v, i)) != cst.root()  ){
		dfs(cst, temp, nodes);
		++i;
	}
}


template<class CST>
void performe(string result_file_name){
	typedef typename CST::size_type size_type;
	typedef typename CST::node_type node_type;
	path test_file_dir("./test_cases/");
	string suffix = ".50MB";
	ofstream result_file(result_file_name.c_str(), ios::out | ios::trunc);
	for(directory_iterator dir_it(test_file_dir), end; dir_it != end; ++dir_it){
		if( is_regular( dir_it->status() ) ){// is regular file in boost 1.36.0
			string test_case = dir_it->path().leaf();	
			string path_name = dir_it->path().file_string();
			string test_case_dir = test_file_dir.directory_string();
			string data_dir = test_case_dir+"../cst_data/";
			if( test_case.size() < suffix.size() )
				continue;
			if( string(test_case, test_case.size()-suffix.size()) == suffix ){


				cerr << test_case << endl;
				result_file << test_case <<endl;
				stop_watch stopwatch;


				{
					string cst_suffix = "";
					CST cst; // generate compressed suffix array
					stringstream ss;
					ss<<"_"<<cst.csa.get_sample_dens()<<"_"<< CST::csa_type::enc_vector_type::sample_dens;
					cst_suffix += ss.str();
					cst_suffix += cst_name_trait<CST>::file_name_suffix();
//					}else
					if( create_or_load_cst(cst, cst_suffix, test_case, test_case_dir, data_dir) ){
						cerr << (int)cst.csa.sigma << " # cst.csa.sigma" << endl;
						result_file << (int)cst.csa.sigma << " # cst.csa.sigma" << endl;
						cerr <<  cst.csa.size() << " # cst.csa.size" << endl;
						result_file <<  cst.csa.size() << " # cst.csa.size" << endl;
						onullstream_type nullstream;
						cerr << cst.serialize(nullstream) << " # size of cst in bytes " << endl;
						result_file << cst.serialize(nullstream) << " # size of cst in bytes " << endl;
						cerr << cst.csa.serialize(nullstream) << " # size of cst.csa in bytes " << endl;
						result_file << cst.csa.serialize(nullstream) << " # size of cst.csa in bytes " << endl;
						cerr << cst.csa.psi.serialize(nullstream) << " # size of cst.csa.psi in bytes" << endl;
						result_file << cst.csa.psi.serialize(nullstream) << " # size of cst.csa.psi in bytes" << endl;
						cerr << cst.lcp.serialize(nullstream) << "  # size of cst.lcp in bytes" <<endl;
						result_file << cst.lcp.serialize(nullstream) << "  # size of cst.lcp in bytes" <<endl;
						cerr << cst.bp.serialize(nullstream) << " # size of cst.bp in bytes" <<endl;
						result_file << cst.bp.serialize(nullstream) << " # size of cst.bp in bytes" <<endl;
						cerr << cst.bp_support.serialize(nullstream) << " # size of cst.bp_support in bytes" << endl;
						result_file << cst.bp_support.serialize(nullstream) << " # size of cst.bp_support in bytes" << endl;
						
						size_type operations = 0;

						size_type extract_len = 80;
						if(extract_len > cst.csa.size())
							extract_len = cst.csa.size();
						string text = algorithm::extract(cst.csa, 0, extract_len);

//						std::cerr<<"Extracted text = "<<std::endl;
//						std::cerr<< text << std::endl;
//						cerr << "#-- runtime performance information " <<endl;

						node_type root = cst.root();
						stopwatch.start();
						for(size_type i=1, j, char_pos; i <= 20000; ++i){
							node_type child = cst.child(root, (unsigned char)text[i%extract_len], char_pos );
							++operations;
							j = i;
							while(child != cst.root() ){
								++j;
								child = cst.child(child, text[(j)%extract_len], char_pos);
								++operations;
							}
						}
						
						stopwatch.stop();
						cerr << stopwatch.getUserTime()/(double)operations << " # ms per child(v,c) operation / ";
						result_file << stopwatch.getUserTime()/(double)operations << " # ms per child(v,c) operation / ";
						cerr << "# " << operations << " child operations take "<<stopwatch.getUserTime()<<" ms"<< endl;
						result_file << "# " << operations << " child operations take "<<stopwatch.getUserTime()<<" ms"<< endl;

						size_type leafs = 60000;
						vector<node_type> leaf(leafs);
						for(size_type i=1; i <= leafs; ++i){
							leaf[i-1] = cst.ith_leaf(i);
						}

						operations = 0;

						stopwatch.start();
						for(size_type i=0; i < leaf.size(); ++i){
							node_type child = cst.parent(leaf[i]);
//							std::cerr<<child<<std::endl;
							++operations;
							while(child != cst.root() ){
								child = cst.parent(child);
								++operations;
							}
						}
						stopwatch.stop();
						cerr << stopwatch.getUserTime()/(double)operations << " # ms per parent(v) operation / ";				
						result_file << stopwatch.getUserTime()/(double)operations << " # ms per parent(v) operation / ";				
						cerr << "# " <<operations << " parent operations take "<<stopwatch.getUserTime()<<" ms"<< endl;
						result_file << "# " <<operations << " parent operations take "<<stopwatch.getUserTime()<<" ms"<< endl;

						cerr<<endl;
						result_file<<endl;
						
/*						
						vector<node_type> nodes;

						for(size_type i=1, size = cst.csa.size(); i <= 1000000; ++i){
							node_type child = cst.parent(leaf[i]);
							while(child != cst.root() ){
								nodes.push_back(child);
								child = cst.parent(child);
							}
						}
										
						node_type n;
						operations = nodes.size();
						stopwatch.start();
						for(size_type i=0; i<nodes.size(); ++i){
							n = cst.sibling(nodes[i]);
							k += n.l;
						}
						stopwatch.stop();
						cerr << operations << " sibling operations take "<<stopwatch.getUserTime()<<" ms"<< endl;
						cerr << stopwatch.getUserTime()/(double)operations << " ms per operation" <<endl;				

						cerr << "k = " << k <<endl;
*/

					}
				}


			}
		}	
	}
	result_file.close();
}


int main(){
//	performe<cst_sct<> >("results/cst_study_sct.txt");
//	performe<cst_sada<> >("results/cst_study_sada.txt");
	typedef csa_sada_prac<enc_vector_prac<coder::fibonacci, 80>, 12, 12> csa_sada_prac_fib_80;
	performe<cst_sct<csa_sada_prac_fib_80> >("results/cst_study_sct_fib_80_12_12.txt");
	performe<cst_sada<csa_sada_prac_fib_80> >("results/cst_study_sada_fib_80_12_12.txt");
	typedef csa_sada_prac<enc_vector_prac<coder::elias_delta, 80>, 12, 12> csa_sada_prac_ed_80;
	performe<cst_sct<csa_sada_prac_ed_80> >("results/cst_study_sct_ed_80_12_12.txt");
	performe<cst_sada<csa_sada_prac_ed_80> >("results/cst_study_sada_ed_80_12_12.txt");
//	performe<cst_sct<cst_sct<>::csa_type, int_vector<0> >	>("results/cst_study_sct_fix_lcp.txt");
//	performe<cst_sada<cst_sada<>::csa_type, int_vector<0> >	>("results/cst_study_sada_fix_lcp.txt");
//	performe<cst_sct<csa_uncompressed, int_vector<0> >	>("results/cst_study_sct_all_fix.txt");
//	performe<cst_sada<csa_uncompressed, int_vector<0> >	>("results/cst_study_sada_all_fix.txt");
}

