#include <boost/filesystem.hpp>
//#include "interface.h" // pizza chili interface

#include <sdsl/testutils.hpp>
#include <sdsl/algorithms.hpp>
#include <sdsl/csa_sada_prac.hpp>
#include <sdsl/cst_sct.hpp>
#include "maekinen/cst_v_1_0/SSTree.h" // maekinen implementation of sadakane cst

#if BOOST_WORKAROUND( __GNUC__, < 3)
typedef boost::test_tools::output_test_stream onullstream_type;
#else
#include <boost/test/utils/nullstream.hpp>
typedef boost::onullstream onullstream_type;
#endif

#include <vector>
#include <set>
#include <string>

using namespace sdsl;
using namespace boost::filesystem;
using namespace std;

typedef cst_sct<> tCST;
//typedef cst_sct<cst_sct<>::csa_type, int_vector<0>, cst_sct<>::bp_support_type > tCST_fix_lcp;
typedef cst_sct<csa_sada_prac<csa_sada_prac<>::enc_vector_type, 2, 2>, int_vector<0>, cst_sct<>::bp_support_type > tCST_fix_lcp;
typedef tCST::node_type node_type;
typedef tCST::size_type size_type;


template <class CST>
bool create_or_load_cst(CST &cst, string cst_suffix, string csa_suffix, string test_case, string test_case_dir, string data_dir){
	typedef typename CST::csa_type CSA;
	if( !util::loadFromFile(cst, (data_dir+test_case+cst_suffix).c_str() ) ){
		char *text = NULL;
		uint64_t n=0;
		if( 0==(n=file::readText((test_case_dir+test_case).c_str(),text)) ){
			cerr << "Could not read file "<<(test_case_dir+test_case)<<". Exiting. "<< endl;	
			return false;
		}	

		CSA csa;
		if( !util::loadFromFile(csa, (data_dir+test_case+csa_suffix).c_str()) ){
			int_vector<32> sa;
			// if we already calculated the suffix array, laod it from disc
			if( !util::loadFromFile(sa, (data_dir+test_case+".sa").c_str() ) ){ 
				sa.resize(n+1);
				algorithm::calculate_sa((const unsigned char*)text,n+1,sa);
				util::storeToFile(sa, ((data_dir+test_case)+".sa").c_str() );
			}
			csa = CSA(sa, (const unsigned char*)text);
			sa.resize(0); // clear sa
			util::storeToFile(csa, (data_dir+test_case+csa_suffix).c_str() );
			cerr<<"store file "<<(data_dir+test_case+csa_suffix).c_str() << endl;
		}
		cst = CST(csa, (const unsigned char*)text);
		util::storeToFile(cst, (data_dir+test_case+cst_suffix).c_str() );
		std::cerr<<"free text"<<std::endl;
		delete [] text;
		std::cerr<<"text was deleted"<<std::endl;
	}
	return true;
}

void dfs(const tCST &cst, node_type v, vector<node_type> &nodes){
	if(nodes.size()%1000000==0){
		std::cerr<<".";
		if(nodes.size()%10000000==0){
			std::cerr<<std::endl;
		}
	}
	nodes.push_back(v);
	size_type i=1;
	node_type temp = cst.root();
	while( (temp = cst.ith_child(v, i)) != cst.root()  ){
		dfs(cst, temp, nodes);
		if(i > cst.csa.sigma ){
			std::cerr<<" more than sigma = "<< (int)cst.csa.sigma <<" children for node "<<v.l<<"-["<<v.left<<","<<v.right<<"]"<<std::endl;
			for(size_type j=1;j<=i;++j){
				temp = cst.ith_child(v, j);
				std::cerr<<setw(3)<<j<<"  "<<temp.l<<"-["<<temp.left<<","<<temp.right<<"]"<<cst.edge(temp, temp.l)<<std::endl;
				if(temp.l < 80){
					for(size_type k=1;k<=temp.l;++k)
						std::cerr<<cst.edge(temp, k);
				}
				std::cerr<<std::endl;
			}
		}
		++i;
	}
}

int main(){
	path test_file_dir("../test_cases/");
	string suffix = ".50MB";
	for(directory_iterator dir_it(test_file_dir), end; dir_it != end; ++dir_it){
		if( is_regular( dir_it->status() ) ){// is regular file in boost 1.36.0
			string test_case = dir_it->path().leaf();	
			string path_name = dir_it->path().file_string();
			string test_case_dir = test_file_dir.directory_string();
			string data_dir = test_case_dir+"../cst_data/";
//			cout<<test_case<<endl;
//			cout<<path_name<<endl;
//			cout << test_case_dir << endl;
//			cout << data_dir << endl;
//			continue;
			if( test_case.size() < suffix.size() )
				continue;
			if( string(test_case, test_case.size()-suffix.size()) == suffix ){

			ofstream result_file;
			size_t entries = 10000000;
			stringstream entriesss;
			entriesss<<entries;
//			result_file.open( (test_case_dir+"../results/"+test_case+"_"+entriesss.str()+".fix.dna.res.log").c_str(), std::ios::out | std::ios::app );
//			if(!result_file){
//				cerr<<"Could not open result logfile! Exit..."<<endl;
//			}

				cerr << test_case << endl;
				stop_watch stopwatch;
				{
					char *text = NULL;
					uint64_t n=0;
					if( 0==(n=file::readText((test_case_dir+test_case).c_str(),text)) ){
						cerr << "Could not read file "<<(test_case_dir+test_case)<<". Exiting. "<< endl;	
						continue;
					}
					std::cerr<<"read text."<<std::endl;
					SSTree *sstree;
					if( boost::filesystem::exists( (data_dir+test_case+".mak.12.cst.csa") ) ){
						std::cerr<<"load cst."<<std::endl;
						sstree = new SSTree((unsigned char*)text, n+1, false, 12, SSTree::load_from, (data_dir+test_case+".mak.12.cst").c_str()); 
					}else{
						std::cerr<<"create cst."<<std::endl;
						sstree = new SSTree((unsigned char*)text, n+1, false, 12, SSTree::save_to, (data_dir+test_case+".mak.12.cst").c_str()); 
					}
					std::cerr<<"cst. was created."<<std::endl;


						size_type operations = 0;

						size_type extract_len = 80;

						typedef uint64_t node_type;

						node_type root = sstree->root();
						stopwatch.start();
						for(size_type i=1, j, char_pos; i <= 100000; ++i){
							node_type child = sstree->child(root, (unsigned char)text[i%extract_len] );
							++operations;
							j = i;
							while(child != 0 ){
								++j;
								child = sstree->child(child, text[(j)%extract_len]);
								++operations;
							}

						}
						
						stopwatch.stop();
						cerr << operations << " child operations take "<<stopwatch.getUserTime()<<" ms"<< endl;
						cerr << stopwatch.getUserTime()/(double)operations << " ms per operation" << endl;


/*						vector<node_type> leaf(cst.csa.size());
						for(size_type i=1; i <= cst.csa.size(); ++i){
							leaf[i-1] = cst.ith_leaf(i);
						}

						stopwatch.start();
						for(size_type i=1, size = cst.csa.size(); i <= 1000000; ++i){
							node_type child = cst.parent(leaf[i]);
							++operations;
							while(child.l > 0 ){
								child = cst.parent(child);
								++operations;
							}
						}
						stopwatch.stop();
						cerr << operations << " parent operations take "<<stopwatch.getUserTime()<<" ms"<< endl;
						cerr << stopwatch.getUserTime()/(double)operations << " ms per operation" <<endl;				
*/
					

					delete sstree;
					delete [] text;
				}
//				cerr << "elias_delta_compression elias_delta_access_time fibonacci_compression fibonacci_access_time sada_compression sada_access_time org_file_size entries sample_psi sample_sa" << endl;
				if(false)
				{// test elias delta implementation of gog
					typedef tCST_fix_lcp CST;
					//typedef tCST CST;
					uint64_t k=0;
					string csa_suffix = "", cst_suffix = "";
//					tCSAEd csa;// generate compressed suffix array
					CST cst;// generate compressed suffix array
					stringstream ss;
					ss<<"_"<<cst.csa.get_sample_dens()<<"_"<< CST::csa_type::enc_vector_type::sample_dens;
					csa_suffix += ss.str();
					cst_suffix += ss.str();
					csa_suffix += ".csa";
//					cst_suffix += ".naive_lcp.sct.cst";
					cst_suffix += ".naive_lcp.sct.cst";
//					}else
					if( create_or_load_cst(cst, cst_suffix, csa_suffix, test_case, test_case_dir, data_dir) ){

						cerr<<"cst.csa.sigma = "<< (int)cst.csa.sigma << endl;
						cerr<<"\"";
						for(size_type i=0; i<cst.csa.sigma; ++i){
							cerr<<cst.csa.comp2char[i];
						}
						cerr<<"\"";
						cerr<<"cst.csa.size  = "<< cst.csa.size() << endl;
						onullstream_type nullstream;
						cerr << cst.serialize(nullstream) << " ";
						cerr << " " << cst.serialize(nullstream)/(double)(cst.size()) << " "<<endl;
						
						for(size_type i=1; i<=cst.csa.sigma; ++i){
							node_type child = cst.ith_child( cst.root(), i );
//							if(child.l > 0)	
								std::cerr<<child;
	//							std::cerr<<cst.edge(child, 1)<<" C["<<i<<"]="<<cst.csa.C[i]<<" "<<child.right-child.left+1<<std::endl;
//							std::cerr<<child.l<<"-["<<child.left<<","<<child.right<<"] "<<std::endl;
							
						}




						size_type operations = 0;

						size_type extract_len = 80;
						if(extract_len > cst.csa.size())
							extract_len = cst.csa.size();
						string text = algorithm::extract(cst.csa, 0, extract_len);

						std::cerr<<"Extracted text = "<<std::endl;
						std::cerr<< text << std::endl;

						node_type root = cst.root();
						stopwatch.start();
						for(size_type i=1, j, char_pos,size = cst.csa.size(); i <= 100000; ++i){
							node_type child = cst.child(root, (unsigned char)text[i%extract_len], char_pos );
//							node_type child1 = cst.child_linear(root, (unsigned char)text[i%extract_len], char_pos);
//							if(child1 != child){
//								std::cerr<<"child  = "<<child<<" child1 = "<<child1<<" "<<std::endl;
//								return 1;
//							}
							++operations;
							j = i;
							while(child.l > 0 ){
								++j;
//								node_type parent = child;
//								child1 = cst.child_linear(child, text[(j)%extract_len], char_pos);
								child = cst.child(child, text[(j)%extract_len], char_pos);
//							if(child1 != child and !cst.is_leaf(child1) ){
//								std::cerr<<"parent = "<<parent<<" char = "<<text[j%extract_len]<<"child  = "<<child<<" child1 = "<<child1<<" "<<std::endl;
//								return 1;
//							}

								++operations;
							}
						}
						
						stopwatch.stop();
						cerr << operations << " child (log sigma) operations take "<<stopwatch.getUserTime()<<" ms"<< endl;
						cerr << stopwatch.getUserTime()/(double)operations << " ms per operation" << endl;

						operations = 0;
						stopwatch.start();
						for(size_type i=1, j, char_pos,size = cst.csa.size(); i <= 100000; ++i){
							node_type child = cst.child_linear(root, (unsigned char)text[i%extract_len], char_pos );
							++operations;
							j = i;
							while(child.l > 0 ){
								child = cst.child_linear(child, text[(++j)%extract_len], char_pos);
								++operations;
							}
						}
						
						stopwatch.stop();
						cerr << operations << " child (sigma) operations take "<<stopwatch.getUserTime()<<" ms"<< endl;
						cerr << stopwatch.getUserTime()/(double)operations << " ms per operation" << endl;

						vector<node_type> leaf(cst.csa.size());
						for(size_type i=1; i <= cst.csa.size(); ++i){
							leaf[i-1] = cst.ith_leaf(i);
						}

						stopwatch.start();
						for(size_type i=1, size = cst.csa.size(); i <= 1000000; ++i){
							node_type child = cst.parent(leaf[i]);
							++operations;
							while(child.l > 0 ){
								child = cst.parent(child);
								++operations;
							}
						}
						stopwatch.stop();
						cerr << operations << " parent operations take "<<stopwatch.getUserTime()<<" ms"<< endl;
						cerr << stopwatch.getUserTime()/(double)operations << " ms per operation" <<endl;				

					}
				}


			}
		}	
	}
}
