extern "C"{
#include "sadakane/comparray4.h" // sadakane's csa
#include "interface.h" // pizza chili interface
}
#include <sdsl/testutils.hpp>
#include <sdsl/enc_vector_prac.hpp>
#include <sdsl/suffixarrays.hpp>

#include <boost/filesystem.hpp>
#if BOOST_WORKAROUND( __GNUC__, < 3)
typedef boost::test_tools::output_test_stream onullstream_type;
#else
#include <boost/test/utils/nullstream.hpp>
typedef boost::onullstream onullstream_type;
#endif

#include <string>

using namespace sdsl;
using namespace boost::filesystem;
using namespace std;

const uint16_t sample_psi = 128;
const uint16_t sample_sa = 16;
typedef csa_sada_prac<enc_vector_prac<coder::fibonacci, sample_psi,32>, sample_sa,32> tCSAFib;
typedef csa_sada_prac<enc_vector_prac<coder::elias_delta, sample_psi,32>, sample_sa, 32> tCSAEd;

//typedef csa_sada_prac<enc_vector_prac<coder::fibonacci, 128>, 16> tCSAFibP;
template <class CSA>
bool create_or_load_csa(CSA &csa, string csa_suffix, string test_case, string test_case_dir, string data_dir){
	if( !util::loadFromFile(csa, (data_dir+test_case+csa_suffix).c_str()) ){
		int_vector<32> sa;
		char *text = NULL;
		// if we already calculated the suffix array, laod it from disc
		if( !util::loadFromFile(sa, (test_case_dir+test_case+".sa").c_str() ) ){ 
			uint64_t n=0;
			if( 0==(n=file::readText((test_case_dir+test_case).c_str(),text)) ){
				cerr << "Could not read file "<<(test_case_dir+test_case)<<". Exiting. "<< endl;	
				return false;
			}	
			sa.resize(n);
			algorithm::calculate_sa((const unsigned char*)text,n,sa);
			util::storeToFile(sa, ((test_case_dir+test_case)+".sa").c_str() );
			delete [] text;
		}else{
		}
		csa = CSA(sa);
		sa.resize(0); // clear sa
		util::storeToFile(csa, (data_dir+test_case+csa_suffix).c_str() );
	}
	return true;
}

int main(){
	path test_file_dir("../test_cases/");
	string suffix = ".50MB";
	for(directory_iterator dir_it(test_file_dir), end; dir_it != end; ++dir_it){
		if( is_regular( dir_it->status() ) ){// is regular file in boost 1.36.0
			string test_case = dir_it->path().leaf();	
			string path_name = dir_it->path().file_string();
			string test_case_dir = test_file_dir.directory_string();
			string data_dir = test_case_dir+"../data/";
//			cout<<test_case<<endl;
//			cout<<path_name<<endl;
//			cout << test_case_dir << endl;
//			cout << data_dir << endl;
//			continue;
			if( test_case.size() < suffix.size() )
				continue;
			if( string(test_case, test_case.size()-suffix.size()) == suffix ){
			ofstream result_file;
			size_t entries = 1;
			stringstream entriesss;
			entriesss<<entries;
//			result_file.open( (path_name+"_"+sdsl::clock::getTimeString()+"_"+".res.log").c_str(), std::ios::out );
			result_file.open( (test_case_dir+"../results/"+test_case+"_"+entriesss.str()+".fix.res.log").c_str(), std::ios::out | std::ios::app );
			if(!result_file){
				cerr<<"Could not open result logfile! Exit..."<<endl;
			}

				cerr << test_case << endl;
				stop_watch stopwatch;

//				cerr << "elias_delta_compression elias_delta_access_time fibonacci_compression fibonacci_access_time sada_compression sada_access_time org_file_size entries sample_psi sample_sa" << endl;

				{// test elias delta implementation of gog
					uint64_t k=0;
					string csa_suffix = "";
//					tCSAEd csa;// generate compressed suffix array
					tCSAFib csa;// generate compressed suffix array
					stringstream ss;
					ss<<"_"<<csa.get_sample_dens()<<"_"<<tCSAEd::enc_vector_type::sample_dens;
					csa_suffix += ss.str();
//					if( typeid(tCSAEd::enc_vector_type::coder) == typeid(elias_delta_coder) ){
//					csa_suffix += ".fix.ed.csa";
					csa_suffix += ".fix.fib.csa";
//					}else
					if( create_or_load_csa(csa, csa_suffix, test_case, test_case_dir, data_dir) ){
						onullstream_type nullstream;
						cerr << csa.serialize(nullstream) << " ";
						cerr << " " << csa.serialize(nullstream)/(double)(csa.size()) << " ";
						
						uint64_t zz=0,last=0, ones=0, lowerthan6766=0, lowerthan16=0;
						const uint16_t max_sum=2048, max_cnt=40;
						uint64_t sums_in_64_bit[max_sum+1]={0}, sum64=0, cnt_in_64_bit[max_cnt+1]={0};
					    double	cnt_in_64_bit_sum[max_cnt+1]={0};
						uint8_t index=0,cnt=0;
						for(tCSAFib::enc_vector_type::const_iterator it = csa.psi.begin(), end=csa.psi.end(); it!=end; ++it,++zz){
							uint64_t v = *it;
							if(false){
								if(v == last+1 )
									++ones;
								if(v < last+6766 )
									++lowerthan6766;
								if(v < last+16)
									++lowerthan16;	
							}
							index += tCSAFib::enc_vector_type::coder::encoding_length(v-last);
							sum64 += (v-last);
							++cnt;
//							cerr<<(int)index<<" "<<sum64<<endl;
//							string s;
//							cin>>s;
							if(index>=64){
								if(index >= 128)
									++cnt_in_64_bit[0];
								if(sum64 >= max_sum)
									++sums_in_64_bit[max_sum];
								else
									++sums_in_64_bit[sum64];
								++cnt_in_64_bit[cnt];
								if(sum64 < 256)
									cnt_in_64_bit_sum[cnt]+=1;
								cnt=0;
								sum64=0;
								index %= 64;
							}
							last = v;
						}

						ofstream statistics_result_file;
						statistics_result_file.open( (test_case_dir+"../results/"+test_case+"_statistics_of_coded_vals.log").c_str(), std::ios::out | std::ios::trunc );
//						for(int i=0; i < max_sum+1; ++i){
//							cerr<<i<<" "<<sums_in_64_bit[i]<<endl;
//							statistics_result_file<<i<<" "<<sums_in_64_bit[i]<<endl;
//						}
						cerr<<endl;
						for(int i=0; i < max_cnt+1; ++i)
							if(cnt_in_64_bit[i])
								cnt_in_64_bit_sum[i] /= cnt_in_64_bit[i]; 
						for(int i=0; i < max_cnt+1; ++i){
							cerr<<i<<" "<<cnt_in_64_bit[i]<<" "<<cnt_in_64_bit_sum[i]<<endl;
							statistics_result_file<<i<<" "<<cnt_in_64_bit[i]<<" "<<cnt_in_64_bit_sum[i]<<endl;
						}
						statistics_result_file.close();
//						cerr << " ones " << ones << "  all "<<zz<<" "<<" ratio "<< ((double)(ones))/zz << endl;
	//					result_file << " ones " << ones << "  all "<<zz<<" "<<" ratio "<< ((double)(ones))/zz << endl;
//						cerr << " lower than 6766 " << lowerthan6766 << "  all "<<zz<<" "<<" ratio "<< ((double)(lowerthan6766))/zz << endl;
//						result_file << " lower than 6766 " << lowerthan6766 << "  all "<<zz<<" "<<" ratio "<< ((double)(lowerthan6766))/zz << endl;
//						result_file << " lower than 16 " << lowerthan16 << "  all "<<zz<<" "<<" ratio "<< ((double)(lowerthan16))/zz << endl;
//						cerr << " lower than 16 " << lowerthan16 << "  all "<<zz<<" "<<" ratio "<< ((double)(lowerthan16))/zz << endl;
						continue;

						result_file << csa.serialize(nullstream)/(double)(csa.size()) << " ";
						cout.flush();
						k=0;
						stopwatch.start();
						for(size_t i=0; i<entries; ++i){
							k += csa[i];
						}
						stopwatch.stop();
						cerr << stopwatch.getUserTime()/(double)entries << " ";
						result_file << stopwatch.getUserTime()/(double)entries << " ";
						cerr << "k=" << k+entries << " "; 
						cerr << "n=" << csa.size() << " ";
					}
				}


				{// test fibonacci implementation of gog
					uint64_t k=0;
					string csa_suffix = "";
					tCSAFib csa;// generate compressed suffix array
					stringstream ss;
					ss<<"_"<<csa.get_sample_dens()<<"_"<<tCSAFib::enc_vector_type::sample_dens;
					csa_suffix += ss.str();
					csa_suffix += ".fix.fib.csa";
					if( create_or_load_csa(csa, csa_suffix, test_case, test_case_dir, data_dir) ){
						onullstream_type nullstream;
						cerr << " " << csa.serialize(nullstream)/(double)(csa.size()) << " ";
						result_file << csa.serialize(nullstream)/(double)(csa.size()) << " ";
						cout.flush();
						k=0;
						stopwatch.start();
						for(size_t i=0; i<entries; ++i){
							k += csa[i];
						}
						stopwatch.stop();
						cerr << stopwatch.getUserTime()/(double)entries << " ";
						result_file << stopwatch.getUserTime()/(double)entries << " ";
						cerr << "k=" << k+entries << " "; 
					}
				}

				if(true)
				{// test implementation of sadakane
					uint64_t n,k=0;
					char *text = NULL;
					char *options = new char[1024];
					stringstream ss;
					ss<<"_"<<sample_sa<<"_"<<sample_psi;
					string sada_path_name = data_dir+test_case+ss.str();
					stringstream optionss;
					optionss<<" samplerate "<<sample_sa<<" samplepsi "<<sample_psi;
					strcpy( options, ("filename "+sada_path_name+optionss.str()).c_str() );
					if( 0==(n=file::readText((test_case_dir+test_case).c_str(),text)) ){
						cerr << "Could not read file "<<path_name<<". Exiting. "<< endl;	
						return 0;
					}else{
//						cerr << "Loaded file " << path_name <<". n = " << n << endl;
					}	
					void **index = (void**)new (CSA*);
					if( !boost::filesystem::exists( sada_path_name+".idx" ) or !boost::filesystem::exists( sada_path_name+".psi" ) ){
//						std::cerr<<sada_path_name+".idx"<<" does not exists. Creating the index...."<<std::endl;
						build_index((uchar*)text, n, options , index);
					}
					else{
//						std::cerr<<"Loading "<<sada_path_name+".idx"<<std::endl;
						char *filename = new char[1024];
						strcpy( filename, sada_path_name.c_str());
						load_index(filename, index);
						delete [] filename;
					}
//					cerr << "Build indx with options " << options << endl;
					delete [] text; delete [] options;
					CSA *sa = (CSA*)*index;
					ulong *size = new ulong;
					index_size(sa, size);
					cerr<<*size/double(n)<<" ";
					result_file << *size/(double)n << " ";
					cout.flush();
					delete size;
					k=0;
					stopwatch.start();
					for(size_t i=1; i<=entries; ++i){
				  		k+=csa_lookup(sa, i);	
					}
					stopwatch.stop();		
					cerr << stopwatch.getUserTime()/(double)entries << " ";
					result_file << stopwatch.getUserTime()/(double)entries << " ";
					cerr << "k=" << k << " "; 
					cerr << "n=" << n << " ";
				  	free_index(*index);
				}
				cerr<< boost::filesystem::file_size( path_name )<<" ";
				result_file << boost::filesystem::file_size( path_name )<<" ";
				cerr << entries << " ";
				result_file << entries << " ";
				cerr << sample_psi << " ";
				result_file << sample_psi << " ";
				cerr << sample_sa << endl;
				result_file << sample_sa << endl;
				result_file.close();
			}
		}	
	}
}
