#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/rank_support.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/testutils.hpp>

using namespace boost::unit_test;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( rank_support_performance_suite )

template<class RankClass>
typename RankClass::size_type calc_sumsum_naive(const bit_vector &b){
	typedef typename RankClass::size_type size_type; 
	size_type result = 0;
	if(typeid(RankClass) != typeid(rank_support_v10) and typeid(RankClass) != typeid(rank_support_v<10,2>) ){
		for(size_type i=1, sum=0, size = b.size(); i <= size; ++i){
			if(b[i-1])
				++sum;
			result += sum;
		}
	}
	else{
		for(size_type i=2, sum=0, size = b.size(); i <= size; ++i){
			if(b[i-1]==0 and b[i-2]==1)
				++sum;
			result += sum;
		}
	}
	
	return result;
}

template<class RankClass>
void performance_test_for_rank(){
	typedef typename RankClass::size_type size_type; 
	std::string class_name = typeid(RankClass).name();

	BOOST_TEST_MESSAGE("Running performance test for "<< class_name <<"...");
	bit_vector v(100000000);
	util::setRandomBits(v);
	bit_vector w(v.size());
	util::setOneBits(w);
	for(size_type i=0; i<w.size()/2;++i)
		w[i]=0;
	
	stop_watch stopWatch;

	stopWatch.start();
	RankClass rs_v(&v);
	stopWatch.stop();
	BOOST_MESSAGE("Creation of "<< class_name <<" for v (size: "<<v.bit_size()<<" bits) takes "<<stopWatch.getUserTime()  << " ms" );
	
	stopWatch.start();
	RankClass rs_w(&w);
	stopWatch.stop();
	BOOST_MESSAGE("Creation of "<< class_name <<" for w (size: "<<v.bit_size()<<" bits) takes "<< stopWatch.getUserTime() << " ms" );

	size_type k = 0, kv=0, kw=0;
	size_type size= v.size();
	stopWatch.start();
	for(size_type i=1; i <= size; ++i){
		k += rs_v.rank(i);
	}
	stopWatch.stop();
	BOOST_MESSAGE("Time for "<<v.size()<<" "<< class_name <<" calls one v "<<stopWatch.getUserTime()  << " ms" );
	BOOST_MESSAGE("Per call: "<<stopWatch.getUserTime() / v.size()  << " ms" );

	kv = calc_sumsum_naive<RankClass>(v);
	BOOST_CHECK( k == kv );

	k=0;
	stopWatch.start();
	for(size_type i=1; i <= size; ++i){
		k += rs_w.rank(i);
	}
	stopWatch.stop();
	BOOST_MESSAGE("Time for "<<v.size()<<" "<< class_name <<" calls one w "<<stopWatch.getUserTime()   << " ms" );
	BOOST_MESSAGE("Per call: "<<stopWatch.getUserTime() / v.size()  << " ms" );
	kw = calc_sumsum_naive<RankClass>(w);
	BOOST_CHECK( k == kw);
}


BOOST_AUTO_TEST_CASE( performance_test )
{
	unit_test_log.set_threshold_level( log_messages );
	performance_test_for_rank<rank_support_jmc>();
	performance_test_for_rank<rank_support_v<> >();
	performance_test_for_rank<rank_support_v<10,2> >();
	performance_test_for_rank<rank_support_v10 >();
}


BOOST_AUTO_TEST_SUITE_END()
