#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/testutils.hpp>

using namespace boost::unit_test;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( select_support_performance_suite )

BOOST_AUTO_TEST_CASE( performance_test )
{
	typedef bit_vector::size_type size_type;
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE("Running performance test for select_support");
	stop_watch stopWatch;
	size_type k=0;

	int_vector<1> v(100000000);
	util::setRandomBits(v);

	stopWatch.start();
	select_support_mcl<1> ss_v(&v);
	stopWatch.stop();

	BOOST_MESSAGE("Creation of select_support<1,1> for v (size: "<<v.bit_size()<<" bits) takes "<< stopWatch.getUserTime() << " ms" );

	size_type args = util::getOneBits(v);

	stopWatch.start();
	for(size_type i=1; i <= args; ++i){
		k += ss_v.select(i);
	}
	stopWatch.stop();
	BOOST_MESSAGE("Time for "<< args << " select_support<1,1>.select calls "<< stopWatch.getUserTime() << " ms" );
	BOOST_MESSAGE("Per call: "<< stopWatch.getUserTime()/args << " ms" );


	size_type k_check = 0; 
	for(size_type i=0; i <= v.size() ;++i){
		if( v[i] )
			k_check += i;
	}
	BOOST_CHECK( k == k_check );


	stopWatch.start();
	select_support_mcl<10,2> ss10_v(&v);
	stopWatch.stop();

	BOOST_MESSAGE("Creation of select_support<10,2> for v (size: "<<v.bit_size()<<" bits) takes "<< stopWatch.getUserTime() << " ms" );

	args = util::getOneZeroBits(v);

	stopWatch.start();
	for(size_type i=1; i <= args; ++i){
		k += ss_v.select(i);
	}
	stopWatch.stop();
	BOOST_MESSAGE("Time for "<< args << " select_support<10,2>.select calls "<< stopWatch.getUserTime() << " ms" );
	BOOST_MESSAGE("Per call: "<< stopWatch.getUserTime()/args << " ms" );


	for(size_type i=0; i<v.size()/3; ++i)
		v[i] = (i%100)==0;
	args = util::getOneBits(v);

	stopWatch.start();
	ss_v.init(&v);
	stopWatch.stop();

	BOOST_MESSAGE("Creation of select_support<1> for v (size: "<<v.bit_size()<<" bits) takes "<< stopWatch.getUserTime() << " ms" );

	k = 0;
	stopWatch.start();
	for(size_type i=1; i <= args; ++i){
		k += ss_v.select(i);
	}
	stopWatch.stop();
	BOOST_MESSAGE("Time for "<< args << " select_support<1>.select calls  "<< stopWatch.getUserTime() << " ms" );
	BOOST_MESSAGE("Per call: "<< stopWatch.getUserTime()/args<< " ms" );

	k_check = 0; 
	for(size_type i=1, j=0; i <= args ;++i, ++j){
		while( v[j]==0 )
			++j;
		k_check += j;
	}
	BOOST_CHECK( k == k_check );
}


BOOST_AUTO_TEST_SUITE_END()

