#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/suffixarrays.hpp>
#include <sdsl/algorithms.hpp>
#include <sdsl/testutils.hpp>
#include <vector>
#include <fstream>
#include <sstream>
#if BOOST_WORKAROUND( __GNUC__, < 3)
typedef boost::test_tools::output_test_stream onullstream_type;
#else
#include <boost/test/utils/nullstream.hpp>
typedef boost::onullstream onullstream_type;
#endif

using namespace boost::unit_test;
using namespace sdsl;

typedef std::vector<size_t> tVI;
typedef csa_sada_theo<> tCSAS;
//typedef csa_sada_prac<enc_vector<coder::fibonacci,16>,10> tCSASL; // with this configuration the vector uses space equal to the space Sadakane's original implementation sada_csa.tar
typedef csa_sada_prac<enc_vector<coder::fibonacci,16>,8> tCSASL;
typedef csa_sada_prac<enc_vector<coder::elias_delta,16>,8> tCSASLE;
typedef csa_sada_prac<enc_vector<coder::ternary,16>,8> tCSASLT;

BOOST_AUTO_TEST_SUITE( SDSCompressedSuffixArray_performance_suite )

BOOST_AUTO_TEST_CASE( performance_test )
{
//	const size_t BLOCK_SIZE = (1<<20); // block size for reading from a file

	onullstream_type dev_null;
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE("Running performance test for csa_sada_theo");

	stop_watch stopwatch;	

	tVI test_size(13);
	test_size[0] = 1000; test_size[1] = 10000; test_size[2] = 100000; test_size[3] = 200000; test_size[4] = 400000;
	test_size[5] = 800000; test_size[6] = 1600000; test_size[7] = 3200000; test_size[8] = 6400000; test_size[9] = 10000000;
	test_size[10] = 25600000; test_size[11] = 51200000; test_size[12] = 102400000;

	BOOST_MESSAGE("Test Compressed Suffix Arrays on Random Data");

	for(size_t tc = 6; tc<7/*test_size.size()*/; ++tc){
		size_t n = test_size[tc];
		BOOST_MESSAGE("n = "<<n);
		char *text = new char[n];
		for(size_t i=0; i<n; ++i)
			text[i] = 'a' + rand()%26;
		text[n-1] = '\0';
		
		tCSASL::value_type k = 0;

		stopwatch.start();
		tCSASL csasl((const unsigned char*)text);
		stopwatch.stop();
		BOOST_MESSAGE("creation of csasl takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		BOOST_MESSAGE("size of csasl "<<csasl.serialize(dev_null) <<" bytes");

		stopwatch.start(); 
		for(tCSASL::size_type i=0, size=csasl.size();i!=size; ++i){
			k+=csasl[i];
		}
		stopwatch.stop();
		BOOST_MESSAGE("iteration of csasl takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");
		BOOST_MESSAGE("k="<<k);
		{
			tCSASL temp;
			temp.swap(csasl);	
		}

		k=0;
		stopwatch.start();
		tCSASLE csasle((const unsigned char*)text);
		stopwatch.stop();
		BOOST_MESSAGE("creation of csasle takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		BOOST_MESSAGE("size of csasle "<<csasle.serialize(dev_null) <<" bytes");
		stopwatch.start(); 
		for(tCSASL::size_type i=0, size=csasle.size();i!=size; ++i){
			k+=csasle[i];
		}
		stopwatch.stop();
		BOOST_MESSAGE("iteration of csasle takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");
		BOOST_MESSAGE("k="<<k);

		{
			tCSASLE temp;
			temp.swap(csasle);	
		}
		k=0;
		stopwatch.start();
		tCSASLT csaslt((const unsigned char*)text);
		stopwatch.stop();
		BOOST_MESSAGE("creation of csaslt takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");
		BOOST_MESSAGE("size of csaslt "<<csaslt.serialize(dev_null) <<" bytes");
		stopwatch.start(); 
		for(tCSASL::size_type i=0, size=csaslt.size();i!=size; ++i){
			k+=csaslt[i];
		}
		stopwatch.stop();
		BOOST_MESSAGE("iteration of csaslt takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");
		BOOST_MESSAGE("k="<<k);

/*
		{  // resize caas
			tCSASL temp;
			csas.swap(temp);
		}

		stopwatch.start();
		tCSASL csas((const char*)text);
		stopwatch.stop();
		BOOST_MESSAGE("creation of csas takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		BOOST_MESSAGE("size of csas "<<csas.serialize(dev_null) <<" bytes");

		stopwatch.start(); 
		for(tCSAS::const_iterator it = csas.begin(), end = csas.end(); it!=end; ++it){
			k1+=*it;	
		}
		stopwatch.stop();
		BOOST_MESSAGE("iteration of csas takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		BOOST_CHECK(k==k1);
		BOOST_MESSAGE("k="<<k);
*/
		delete text;
	}

	char *c = NULL;
	if( file::readText("performance/english.50MB", c) ){
		BOOST_MESSAGE("Creating csasl from text");
		stopwatch.start();
		tCSASL csasl((const unsigned char*)c);
		stopwatch.stop();
		BOOST_MESSAGE("creation of csasl takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		delete [] c;

		tCSAS::value_type k = 0, k1=0;

		BOOST_MESSAGE("size of csasl "<<csasl.serialize(dev_null) <<" bytes");
/*		stopwatch.start(); 
		for(tCSASL::const_iterator it = csasl.begin(), end = csasl.end(); it!=end; ++it){
			k+=*it;	
		}
		stopwatch.stop();
		BOOST_MESSAGE("iteration of csasl takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");
*/
		stopwatch.start(); 
		for(size_t i=0, end = csasl.size(); i<end; ++i){
			k1+=csasl[i];	
		}
		stopwatch.stop();
		BOOST_MESSAGE("iteration of csasl takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		BOOST_CHECK(k==k1);
/*		stopwatch.start();
		tCSASL csasl(csas);
		stopwatch.stop();

		{
			tCSAS temp;
			csas.swap(temp);
		}

		BOOST_MESSAGE("creation of csasl from csas takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		stopwatch.start();
		tCSAS csas2(csasl);
		stopwatch.stop();

		BOOST_MESSAGE("creation of csas2 from csasl takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		stopwatch.start();
		BOOST_CHECK( csas == csas2 );
		stopwatch.stop();

		BOOST_MESSAGE("csas == csas2  takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");


		stopwatch.start();
		BOOST_MESSAGE( SDSAlgorithm::equal_container_values(csas, csas2) );
		stopwatch.stop();

		BOOST_MESSAGE("equal_container_values(csas, csas2)  takes "<<stopwatch.getUserTime()<<" ms usr "<<stopwatch.getUserTime()+stopwatch.getSysTime()<<" ms usr+sys "<<stopwatch.getRealTime()<<" ms real time.");

		{tCSAS temp;
			temp.swap(csas);
		}	
*/
	}else{
		BOOST_MESSAGE("Could not load english text test files. Please run the performance test from the base directory.");
	}


}

BOOST_AUTO_TEST_SUITE_END()
