#include <sdsl/testutils.hpp>
#include <sdsl/enc_vector_prac.hpp>
#include <sdsl/enc_vector_dna.hpp>
#include <sdsl/suffixarrays.hpp>

#include <boost/filesystem.hpp>
#if BOOST_WORKAROUND( __GNUC__, < 3)
typedef boost::test_tools::output_test_stream onullstream_type;
#else
#include <boost/test/utils/nullstream.hpp>
typedef boost::onullstream onullstream_type;
#endif

#include <string>

using namespace sdsl;
using namespace boost::filesystem;
using namespace std;

const uint32_t sample_psi = 256;
const uint16_t sample_sa = 16;
typedef csa_sada_prac<enc_vector_prac<coder::fibonacci, sample_psi,32>, sample_sa,32> tCSAFib;
typedef csa_sada_prac<enc_vector_dna<sample_psi,32>, sample_sa,32> tCSADNA;
typedef csa_sada_prac<enc_vector_prac<coder::elias_delta, sample_psi,32>, sample_sa, 32> tCSAEd;

//typedef csa_sada_prac<enc_vector_prac<coder::fibonacci, 128>, 16> tCSAFibP;
template <class CSA>
bool create_or_load_csa(CSA &csa, string csa_suffix, string test_case, string test_case_dir, string data_dir){
	if( !util::loadFromFile(csa, (data_dir+test_case+csa_suffix).c_str()) ){
		int_vector<32> sa;
		char *text = NULL;
		// if we already calculated the suffix array, laod it from disc
		if( !util::loadFromFile(sa, (test_case_dir+test_case+".sa").c_str() ) ){ 
			uint64_t n=0;
			if( 0==(n=file::readText((test_case_dir+test_case).c_str(),text)) ){
				cerr << "Could not read file "<<(test_case_dir+test_case)<<". Exiting. "<< endl;	
				return false;
			}	
			sa.resize(n);
			algorithm::calculate_sa((const unsigned char*)text,n,sa);
			util::storeToFile(sa, ((test_case_dir+test_case)+".sa").c_str() );
			delete [] text;
		}else{
		}
		csa = CSA(sa);
		sa.resize(0); // clear sa
		util::storeToFile(csa, (data_dir+test_case+csa_suffix).c_str() );
		cerr<<"store file "<<(data_dir+test_case+csa_suffix).c_str() << endl;
	}
	return true;
}

int main(){
	path test_file_dir("./test_cases/");
	string suffix = "dna.50MB";
	for(directory_iterator dir_it(test_file_dir), end; dir_it != end; ++dir_it){
		if( is_regular( dir_it->status() ) ){// is regular file in boost 1.36.0
			string test_case = dir_it->path().leaf();	
			string path_name = dir_it->path().file_string();
			string test_case_dir = test_file_dir.directory_string();
			string data_dir = test_case_dir+"../csa_data/";
//			cout<<test_case<<endl;
//			cout<<path_name<<endl;
//			cout << test_case_dir << endl;
//			cout << data_dir << endl;
//			continue;
			if( test_case.size() < suffix.size() )
				continue;
			if( string(test_case, test_case.size()-suffix.size()) == suffix ){
			ofstream result_file;
			size_t entries = 10000000;
			stringstream entriesss;
			entriesss<<entries;
//			result_file.open( (path_name+"_"+sdsl::clock::getTimeString()+"_"+".res.log").c_str(), std::ios::out );
			result_file.open( (test_case_dir+"../results/"+test_case+"_"+entriesss.str()+".fix.dna.res.log").c_str(), std::ios::out | std::ios::app );
			if(!result_file){
				cerr<<"Could not open result logfile! Exit..."<<endl;
			}

				cerr << test_case << endl;
				stop_watch stopwatch;

//				cerr << "elias_delta_compression elias_delta_access_time fibonacci_compression fibonacci_access_time sada_compression sada_access_time org_file_size entries sample_psi sample_sa" << endl;

				{// test elias delta implementation of gog
					uint64_t k=0;
					string csa_suffix = "";
//					tCSAEd csa;// generate compressed suffix array
					tCSADNA csa;// generate compressed suffix array
					stringstream ss;
					ss<<"_"<<csa.get_sample_dens()<<"_"<<tCSAEd::enc_vector_type::sample_dens;
					csa_suffix += ss.str();
//					if( typeid(tCSAEd::enc_vector_type::coder) == typeid(elias_delta_coder) ){
//					csa_suffix += ".fix.ed.csa";
					csa_suffix += ".fix.dna.csa";
//					}else
					if( create_or_load_csa(csa, csa_suffix, test_case, test_case_dir, data_dir) ){
						onullstream_type nullstream;
						cerr << csa.serialize(nullstream) << " ";
						cerr << " " << csa.serialize(nullstream)/(double)(csa.size()) << " ";
						

						result_file << csa.serialize(nullstream)/(double)(csa.size()) << " ";
						cout.flush();
						k=0;
						stopwatch.start();
						for(size_t i=0; i<entries; ++i){
							k += csa[i];
						}
						stopwatch.stop();
						cerr << stopwatch.getUserTime()/(double)entries << " ";
						result_file << stopwatch.getUserTime()/(double)entries << " ";
						cerr << "k=" << k+entries << " "; 
						cerr << "n=" << csa.size() << " ";
						cerr << endl;
					}
				}


	continue;
				{// test fibonacci implementation of gog
					uint64_t k=0;
					string csa_suffix = "";
					tCSAFib csa;// generate compressed suffix array
					stringstream ss;
					ss<<"_"<<csa.get_sample_dens()<<"_"<<tCSAFib::enc_vector_type::sample_dens;
					csa_suffix += ss.str();
					csa_suffix += ".fix.fib.csa";
					if( create_or_load_csa(csa, csa_suffix, test_case, test_case_dir, data_dir) ){
						onullstream_type nullstream;
						cerr << " " << csa.serialize(nullstream)/(double)(csa.size()) << " ";
						result_file << csa.serialize(nullstream)/(double)(csa.size()) << " ";
						cout.flush();
						k=0;
						stopwatch.start();
						for(size_t i=0; i<entries; ++i){
							k += csa[i];
						}
						stopwatch.stop();
						cerr << stopwatch.getUserTime()/(double)entries << " ";
						result_file << stopwatch.getUserTime()/(double)entries << " ";
						cerr << "k=" << k+entries << " "; 
						cerr << endl;
					}
				}

				cerr<< boost::filesystem::file_size( path_name )<<" ";
				result_file << boost::filesystem::file_size( path_name )<<" ";
				cerr << entries << " ";
				result_file << entries << " ";
				cerr << sample_psi << " ";
				result_file << sample_psi << " ";
				cerr << sample_sa << endl;
				result_file << sample_sa << endl;
				result_file.close();
			}
		}	
	}
}
